open Format
open Instr
open Types
open Init

module GUI_Update : sig
  type t
  val save : pos -> unit
  val exec : unit -> unit
end = 
struct
  open Gui
  module S = Set.Make(struct type t = pos let compare = compare end)
  type t = S.t
  let tbl = ref S.empty
  let save pos = tbl := S.add pos !tbl
  let exec () =
    S.iter draw_cell !tbl;
    refresh ();
    tbl := S.empty
end

(* subsection Geography *)

let even x = x mod 2 = 0

let adjacent_cell (p:pos) (d:dir) : pos = 
  let (x,y) = p in
  match d with
    | 0 -> (x+1, y)
    | 1 -> if even y  then (x, y+1) else (x+1, y+1)
    | 2 -> if even y then (x-1, y+1) else (x, y+1)
    | 3 -> (x-1, y)
    | 4 -> if even y then (x-1, y-1) else (x, y-1)
    | 5 -> if even y then (x, y-1) else (x+1, y-1)
    | _ -> assert false

let turn (lr:left_or_right) (d:dir) : dir = 
  match lr with
    | Left -> if d=0 then 5 else d-1 (* (d+5) mod 6 *)
    | Right -> if d=5 then 0 else d+1 (* (d+1) mod 6 *)

let sensed_cell (p:pos) (d:dir) (sd:sense_dir) : pos = 
  match sd with
    | Here ->  p
    | Ahead -> adjacent_cell p d
    | LeftAhead -> adjacent_cell p (turn Left d)
    | RightAhead -> adjacent_cell p (turn Right d)

let other_color (c:color) : color = 
  match c with
    | Red -> Black
    | Black -> Red

(*
function state(a:ant):int     = &lt;get state component of a&gt;
function color(a:ant):color   = &lt;get color component of a&gt;
function resting(a:ant):int   = &lt;get resting component of a&gt;
function direction(a:ant):dir = &lt;get direction component of a&gt;
function has_food(a:ant):bool = &lt;get has_food component of a&gt;

function set_state(a:ant, s:int)     = &lt;set state component of a to be s&gt;
function set_resting(a:ant, r:int)   = &lt;set resting component of a to be r&gt;
function set_direction(a:ant, d:dir) = &lt;set direction component of a to be d&gt;
function set_has_food(a:ant, b:bool) = &lt;set has_food component of a to be b&gt;
*)

let rocky (p:pos) : bool = 
  (* true if the cell at position p is rocky *)
  let (x,y) = p in world.(x).(y).rocky
 
let some_ant_is_at (p:pos) : bool =
  (* true if there is an ant in the cell at position p *)
  let (x,y) = p in
  match world.(x).(y).cell_ant with
    | None -> false
    | Some _ -> true

let ant_at (p:pos) : ant =
  (* return the ant in the cell at position p *)
  let (x,y) = p in
  match world.(x).(y).cell_ant with
    | None -> assert false
    | Some a -> ants.(a)
  
let set_ant_at (p:pos) (a:ant) =
  (* record the fact that the given ant is at position p *)
  let (x,y) = p in
  world.(x).(y).cell_ant <- Some a.id;
  a.position <- p
  
let clear_ant_at (p:pos) =
  (* record the fact that no ant is at position p *)
  let (x,y) = p in
   world.(x).(y).cell_ant <- None

let ant_is_alive (id:int) : bool =
  (* true if an ant with the given id exists somewhere in the world *)
  ants.(id).is_alive

let find_ant (id:int) : pos =
  (* return current position of the ant with the given id *)
  ants.(id).position

let kill_ant_at (p:pos) = 
  clear_ant_at p
  
let food_at (p:pos) : int =
  (* return the amount of food in the cell at position p *)
  let (x,y) = p in
   world.(x).(y).food
  
let set_food_at (p:pos) (f:int) =
  (* record the fact that a given amount of food is at position p *)
  let (x,y) = p in
  world.(x).(y).food <- f
  
let anthill_at (p:pos) (c:color) : bool =
  (* true if the cell at position p is in the anthill of color c *)
  let (x,y) = p in
  match world.(x).(y).ant_hill with
    | None -> false
    | Some c' -> c=c'

(* subsection Cartography *)

let fprintf_color fmt c =
  match c with
    | Red -> fprintf fmt "red"
    | Black -> fprintf fmt "black"

let check_any_marker m =
  m.(0) || m.(1) || m.(2) || m.(3) || m.(4) || m.(5)

let dump_marks c marks =
  if check_any_marker marks then
    begin
      printf "%a marks: " fprintf_color c;
      for i=0 to 5 do
	if marks.(i) then printf "%d" i;
      done;
      printf "; "
    end

let dump_world round =
  printf "After round %d...@." round;
  for y=0 to height - 1 do
    for x=0 to width - 1 do
      printf "cell (%d, %d): " x y;
      let cell = world.(x).(y) in
      if cell.rocky then printf "rock" else
	begin
	  if cell.food > 0 then printf "%d food; " cell.food;
	  begin
	    match cell.ant_hill with
	      | None -> () 
	      | Some c -> printf "%a hill; " fprintf_color c
	  end;
	  dump_marks Red cell.red_markers;
	  dump_marks Black cell.black_markers;
	  begin
	    match cell.cell_ant with
	      | None -> () 
	      | Some id -> 
		  let a = ants.(id) in
		  printf "%a ant of id %d, " fprintf_color a.color id;
		  printf "dir %d, food %d, " a.direction (if a.has_food then 1 else 0);
		  printf "state %d, resting %d" a.state a.resting;
	  end;
	end;
      printf "@\n"
    done
  done;
  printf "@\n"


(* subsection Chemistry *)

let set_marker_at (p:pos) (c:color) (i:marker) =
  (* set marker i of color c in cell p *)
  let (x,y) = p in
  match c with
    | Red -> world.(x).(y).red_markers.(i) <- true
    | Black -> world.(x).(y).black_markers.(i) <- true
  
let clear_marker_at (p:pos) (c:color) (i:marker) =
  (* clear marker i of color c in cell p *)
  let (x,y) = p in
  match c with
    | Red -> world.(x).(y).red_markers.(i) <- false
    | Black -> world.(x).(y).black_markers.(i) <- false

let check_marker_at (p:pos) (c:color) (i:marker) : bool =
  (* true if marker i of color c is set in cell p *)
  let (x,y) = p in
  match c with
    | Red -> world.(x).(y).red_markers.(i)
    | Black -> world.(x).(y).black_markers.(i)

let check_any_marker_at (p:pos) (c:color) : bool =
  (* true if ANY marker of color c is set in cell p *)
  let (x,y) = p in
  match c with
    | Red -> check_any_marker world.(x).(y).red_markers
    | Black -> check_any_marker world.(x).(y).black_markers
;;

(* subsection Phenomenology *)

let cell_matches (p:pos) (cond:condition) (c:color) : bool = 
  if rocky p 
  then cond = Rock
  else
    match cond with
      | Friend ->
          some_ant_is_at p &&
          (ant_at p).color = c
      | Foe ->
          some_ant_is_at p &&
          (ant_at p).color <> c
      | FriendWithFood ->
          some_ant_is_at p &&
          (ant_at p).color = c &&
	  (ant_at p).has_food
      | FoeWithFood ->
          some_ant_is_at p &&
          (ant_at p).color <> c &&
          (ant_at p).has_food
      | Food ->
          food_at p > 0
      | Rock ->
          false
      | Marker(i) ->
          check_marker_at p c i
      | FoeMarker ->
          check_any_marker_at p (other_color c)
      | Home ->
          anthill_at p c
      | FoeHome ->
          anthill_at p (other_color c)

(* subsection Neurology *)

let get_instruction (c:color) (s:state) : instruction =
  (* get the instruction for state s of ant color c *)
  match c with
    | Red -> red_brain.(s)
    | Black -> black_brain.(s)

(* subsection Neuro-Cartography *)
  
let dump_brain n b =
  for i=0 to n do
    let ins = b.(i) in
    match ins with
      | Sense(sense_dir, st1, st2, condition) ->
	  eprintf "Sense(...)@."
      | Mark(marker, st) ->
	  eprintf "Mark(...)@."
      | Unmark(marker, st) ->
	  eprintf "Ummark(...)@."
      | PickUp(st1, st2) ->
	  eprintf "PickUp(...)@."
      | Drop(st) ->
	  eprintf "Drop(...)@."
      | Turn(left_or_right, st) ->
	  eprintf "Turn(...)@."
      | Move(st1, st2) ->
	  eprintf "Move(...)@."
      | Flip(n,st1,st2) ->
	  eprintf "Flip %d %d %d@." n st1 st2
  done

(* subsection Martial Arts *)

let adjacent_ants (p:pos) (c:color) : int =
  let n = ref 0 in
  for d = 0 to 5 do
    let cel = adjacent_cell p  d in
    if some_ant_is_at cel && (ant_at cel).color = c 
    then incr n
  done;
  !n

let check_for_surrounded_ant_at (p:pos) =
  if some_ant_is_at p then 
    let a = ant_at p in
    if adjacent_ants p (other_color a.color) >= 5 then 
      begin
	kill_ant_at p;
	a.is_alive <- false;
	set_food_at p (food_at p + 3 + (if a.has_food then 1 else 0))
      end

let check_for_surrounded_ants (p:pos) =
  check_for_surrounded_ant_at p;
  for d = 0 to 5 do
    check_for_surrounded_ant_at (adjacent_cell p d)
  done

(* subsection Number Theory *)

let seed = ref 0

let next_seed () =
  seed := !seed * 22695477 + 1

let randomint (n:int) : int = 
  next_seed ();
  let x = (!seed lsr 16) land 16383 in
  x mod n

let randomseed n =
  seed := n;
  next_seed ();
  next_seed ();
  next_seed ()

let debug_random = false  

let _ =
  if debug_random then
    begin
      randomseed 12345;
      for i=0 to 99 do
	printf "%d, " (randomint 16384);
      done;
      printf "@."
    end

(* subsection Kinetics *)

let step id =
  if ant_is_alive id then
    let p = find_ant id in
    let a = ant_at p in
    if a.resting > 0 then
      a.resting <- a.resting - 1
    else
      match get_instruction a.color a.state with
        | Sense(sensedir, st1, st2, cond) ->
            let p' = sensed_cell p a.direction sensedir in
            let st = if cell_matches p' cond a.color then st1 else st2 in
            a.state <- st
        | Mark(i, st) ->
            set_marker_at p a.color i;
            a.state <- st
        | Unmark(i, st) ->
            clear_marker_at p a.color i;
            a.state <- st
        | PickUp(st1, st2) ->
            if a.has_food || food_at p = 0 then 
              a.state <- st2 
            else begin
              set_food_at p (food_at p - 1);
              a.has_food <- true;
              a.state <- st1;
	      GUI_Update.save p
            end;
        | Drop(st) ->
            if a.has_food then begin
              set_food_at p (food_at p + 1);
              a.has_food <- false;
	      GUI_Update.save p
            end;
            a.state <- st;
        | Turn(lr, st) -> 
            a.direction <- turn lr a.direction;
            a.state <- st;
	    GUI_Update.save p
        | Move(st1, st2) ->
            let newp = adjacent_cell p a.direction in
            if rocky(newp) || some_ant_is_at(newp) then
              a.state <- st2
            else begin
              clear_ant_at p;
              set_ant_at newp a;
              a.state <- st1;
              a.resting <- 14;
              check_for_surrounded_ants(newp)
            end;
	    GUI_Update.save p;
	    GUI_Update.save newp
        | Flip(p, st1, st2) ->
            let st = if randomint p = 0 then st1 else st2 in
            a.state <- st

let round r =
  if dump && r <= 10000 then dump_world r;
  if verbose && r mod 5000 = 0 then eprintf "round %d@." r;
  if r mod speed = 0 then GUI_Update.exec ();
  for id = 0 to List.length ant_hills - 1 do step id done

(* subsection Game Play and Scoring *)

let game () =
  let s = 12345 in
  randomseed s;
  if dump then printf "random seed: %d@.@." s;
  for i = 0 to 99999 do round i done;
  (* score *)
  List.fold_left
    (fun (r,b) (x,y,c) ->
       let cell = world.(x).(y) in
       match c with
	 | Red -> (r + cell.food,b)
	 | Black -> (r, b + cell.food))
    (0,0) ant_hills
;;

(*
    
<LI>At the end, we count the number of food particles currently
 in the anthill cells of each color. The color with the most
 food wins. Remember that food being carried by an ant does not count, even if it 
 is standing on its own anthill.
</UL>
*)

let () = 
  Gui.start ~enter:(fun () -> ignore (game ())) ()
