open Format
open Instr

type pos = int * int

type dir = int (* 0..5 *)

let even x = x mod 2 = 0

let adjacent_cell (p:pos) (d:dir) : pos = 
  let (x,y) = p in
  match d with
    | 0 -> (x+1, y)
    | 1 -> if even y  then (x, y+1) else (x+1, y+1)
    | 2 -> if even y then (x-1, y+1) else (x, y+1)
    | 3 -> (x-1, y)
    | 4 -> if even y then (x-1, y-1) else (x, y-1)
    | 5 -> if even y then (x, y-1) else (x+1, y-1)
    | _ -> assert false

let turn (lr:left_or_right) (d:dir) : dir = 
  match lr with
    | Left -> if d=0 then 5 else d-1 (* (d+5) mod 6 *)
    | Right -> if d=5 then 0 else d+1 (* (d+1) mod 6 *)

let sensed_cell (p:pos) (d:dir) (sd:sense_dir) : pos = 
  match sd with
    | Here ->  p
    | Ahead -> adjacent_cell p d
    | LeftAhead -> adjacent_cell p (turn Left d)
    | RightAhead -> adjacent_cell p (turn Right d)

type color = Red | Black

let other_color (c:color) : color = 
  match c with
    | Red -> Black
    | Black -> Red

type ant = {
  id : int; 
  (* A unique integer that determines the order in which the ant takes its
     turn to move or sense during each round of gameplay. *)

  mutable color : color;

  mutable state : int ;
  (* An integer between 0 and 9999 representing the current
     state of its brain. (All the ants of each species run exactly the
     same program, so the entire neural state of an ant of the species can
     be characterized by just one number. *)

  mutable resting : int ;
  (* An integer that keeps track of how long the ant has to
 rest after its last move before any other action. (Ants move much
 more slowly than they can turn and sense their surroundings. This
 is modeled by making an ant rest for 14 rounds after each time it
 executes a Move instruction. *)

  mutable position : pos ; 

  mutable direction : dir ;

  mutable has_food : bool ;
  (* A boolean indicating whether the ant is currently
     carrying a food particle. (An ant can hold only a single unit of
     food at a time. *)

  mutable is_alive : bool ;
}

(*
function state(a:ant):int     = &lt;get state component of a&gt;
function color(a:ant):color   = &lt;get color component of a&gt;
function resting(a:ant):int   = &lt;get resting component of a&gt;
function direction(a:ant):dir = &lt;get direction component of a&gt;
function has_food(a:ant):bool = &lt;get has_food component of a&gt;

function set_state(a:ant, s:int)     = &lt;set state component of a to be s&gt;
function set_resting(a:ant, r:int)   = &lt;set resting component of a to be r&gt;
function set_direction(a:ant, d:dir) = &lt;set direction component of a to be d&gt;
function set_has_food(a:ant, b:bool) = &lt;set has_food component of a to be b&gt;
*)

let max_ant_id = ref 0

let ants = ref [||]

(* subsection Geography *)

type cell = {
  mutable rocky : bool ;
  mutable cell_ant : int option ; (* the id of the ant there if any *)
  mutable ant_hill : color option ;
  mutable food : int ;
  mutable red_markers : bool array;
  mutable black_markers : bool array;
}

let world_width = ref 0
		    
let world_height = ref 0
		     
let map = ref [||]

let rocky (p:pos) : bool = 
  (* true if the cell at position p is rocky *)
  let (x,y) = p in !map.(x).(y).rocky
 
let some_ant_is_at (p:pos) : bool =
  (* true if there is an ant in the cell at position p *)
  let (x,y) = p in
  match !map.(x).(y).cell_ant with
    | None -> false
    | Some _ -> true

let ant_at (p:pos) : ant =
  (* return the ant in the cell at position p *)
  let (x,y) = p in
  match !map.(x).(y).cell_ant with
    | None -> assert false
    | Some a -> !ants.(a)
  
let set_ant_at (p:pos) (a:ant) =
  (* record the fact that the given ant is at position p *)
  let (x,y) = p in
  !map.(x).(y).cell_ant <- Some a.id;
  a.position <- p
  
let clear_ant_at (p:pos) =
  (* record the fact that no ant is at position p *)
  let (x,y) = p in
   !map.(x).(y).cell_ant <- None

let ant_is_alive (id:int) : bool =
  (* true if an ant with the given id exists somewhere in the world *)
  !ants.(id).is_alive

let find_ant (id:int) : pos =
  (* return current position of the ant with the given id *)
  !ants.(id).position

let kill_ant_at (p:pos) = 
  clear_ant_at p
  

let food_at (p:pos) : int =
  (* return the amount of food in the cell at position p *)
  let (x,y) = p in
   !map.(x).(y).food
  
let set_food_at (p:pos) (f:int) =
  (* record the fact that a given amount of food is at position p *)
  let (x,y) = p in
  !map.(x).(y).food <- f
  
let anthill_at (p:pos) (c:color) : bool =
  (* true if the cell at position p is in the anthill of color c *)
  let (x,y) = p in
  match !map.(x).(y).ant_hill with
    | None -> false
    | Some c' -> c=c'

(* subsection Cartography *)

let rec input_nonspace_char ch =
  let c = input_char ch in
  if c=' ' then input_nonspace_char ch else c

let read_world filename =
  let ch = open_in filename in
  let width = int_of_string (input_line ch) in
  let height = int_of_string (input_line ch) in
  world_width := width;
  world_height := height;
  map := Array.create_matrix width height 
    { rocky = false ; cell_ant = None ; ant_hill = None; food = 0 ;
      red_markers = [||] ; black_markers = [||] };
  let ant_hills = ref [] in
  for y = 0 to height - 1 do
    for x = 0 to width - 1 do
      !map.(x).(y) <-
      { rocky = false ; cell_ant = None ; ant_hill = None; food = 0 ;
	red_markers = [||] ; black_markers = [||] };      
      let c = input_nonspace_char ch in
      begin
	match c with
	| '#' -> !map.(x).(y).rocky <- true 
	| '.' -> ()
	| '+' -> 
	    !map.(x).(y).ant_hill <- Some Red;
	    ant_hills := (x,y,Red):: !ant_hills
	| '-' -> 
	    !map.(x).(y).ant_hill <- Some Black;
	    ant_hills := (x,y,Black):: !ant_hills
	| '1'..'9' -> !map.(x).(y).food <- int_of_char c - int_of_char '0'
	| _ -> 
	    printf "Unknown character %c.@." c
      end;
      if c <> '#' then
	begin
	  !map.(x).(y).red_markers <- Array.create 6 false;
	  !map.(x).(y).black_markers <- Array.create 6 false
	end;
    done;
    let c = input_nonspace_char ch in
    assert (c = '\n')
  done;
  close_in ch;
  !ant_hills

let fprintf_color fmt c =
  match c with
    | Red -> fprintf fmt "red"
    | Black -> fprintf fmt "black"

let check_any_marker m =
  m.(0) || m.(1) || m.(2) || m.(3) || m.(4) || m.(5)

let dump_marks c marks =
  if check_any_marker marks then
    begin
      printf "%a marks: " fprintf_color c;
      for i=0 to 5 do
	if marks.(i) then printf "%d" i;
      done;
      printf "; "
    end

let dump_world round =
  printf "After round %d...@." round;
  for y=0 to !world_height - 1 do
    for x=0 to !world_width - 1 do
      printf "cell (%d, %d): " x y;
      let cell = !map.(x).(y) in
      if cell.rocky then printf "rock" else
	begin
	  if cell.food > 0 then printf "%d food; " cell.food;
	  begin
	    match cell.ant_hill with
	      | None -> () 
	      | Some c -> printf "%a hill; " fprintf_color c
	  end;
	  dump_marks Red cell.red_markers;
	  dump_marks Black cell.black_markers;
	  begin
	    match cell.cell_ant with
	      | None -> () 
	      | Some id -> 
		  let a = !ants.(id) in
		  printf "%a ant of id %d, " fprintf_color a.color id;
		  printf "dir %d, food %d, " a.direction (if a.has_food then 1 else 0);
		  printf "state %d, resting %d" a.state a.resting;
	  end;
	end;
      printf "@\n"
    done
  done;
  printf "@\n"


(* subsection Chemistry *)

let set_marker_at (p:pos) (c:color) (i:marker) =
  (* set marker i of color c in cell p *)
  let (x,y) = p in
  match c with
    | Red -> !map.(x).(y).red_markers.(i) <- true
    | Black -> !map.(x).(y).black_markers.(i) <- true
  
let clear_marker_at (p:pos) (c:color) (i:marker) =
  (* clear marker i of color c in cell p *)
  let (x,y) = p in
  match c with
    | Red -> !map.(x).(y).red_markers.(i) <- false
    | Black -> !map.(x).(y).black_markers.(i) <- false

let check_marker_at (p:pos) (c:color) (i:marker) : bool =
  (* true if marker i of color c is set in cell p *)
  let (x,y) = p in
  match c with
    | Red -> !map.(x).(y).red_markers.(i)
    | Black -> !map.(x).(y).black_markers.(i)

let check_any_marker_at (p:pos) (c:color) : bool =
  (* true if ANY marker of color c is set in cell p *)
  let (x,y) = p in
  match c with
    | Red -> check_any_marker !map.(x).(y).red_markers
    | Black -> check_any_marker !map.(x).(y).black_markers
;;

(* subsection Phenomenology *)

let cell_matches (p:pos) (cond:condition) (c:color) : bool = 
  if rocky p 
  then cond = Rock
  else
    match cond with
      | Friend ->
          some_ant_is_at p &&
          (ant_at p).color = c
      | Foe ->
          some_ant_is_at p &&
          (ant_at p).color <> c
      | FriendWithFood ->
          some_ant_is_at p &&
          (ant_at p).color = c &&
	  (ant_at p).has_food
      | FoeWithFood ->
          some_ant_is_at p &&
          (ant_at p).color <> c &&
          (ant_at p).has_food
      | Food ->
          food_at p > 0
      | Rock ->
          false
      | Marker(i) ->
          check_marker_at p c i
      | FoeMarker ->
          check_any_marker_at p (other_color c)
      | Home ->
          anthill_at p c
      | FoeHome ->
          anthill_at p (other_color c)

(* gra *)

open Graphics

let grey = rgb 128 128 128
let green9 = rgb 128 255 0
let green8 = rgb 128 245 0
let green7 = rgb 128 235 0
let green6 = rgb 128 225 0
let green5 = rgb 128 215 0
let green4 = rgb 128 205 0
let green3 = rgb 128 195 0
let green2 = rgb 128 185 0
let green1 = rgb 128 175 0

let scale_x = 5
let scale_y = 3

let gra_pos x y = 
  let xx= scale_x*(2*x+(y mod 2) - 1) in
  let yy= 900 - scale_y*(3*y+5) in
  (xx,yy)

let show_cell x y =
  let cell = !map.(x).(y) in
  let col =
    if cell.rocky then grey
    else
      match cell.ant_hill with
	| None -> yellow
	| Some Red -> red
	| Some Black -> black
  in
  let (xx,yy) = gra_pos x y in
  set_color col;
  fill_poly [|xx,yy ; 
	      xx,yy+2*scale_y ;
	      xx+scale_x , yy + 3*scale_y ;
	      xx+2*scale_x , yy + 2*scale_y ;
	      xx+2*scale_x , yy ;
	      xx+scale_x , yy - scale_y |];
  if cell.food >0 then
    begin
      let c =
	match cell.food with
	  | 0 -> yellow
	  | 1 -> green1
	  | 2 -> green2
	  | 3 -> green3
	  | 4 -> green4
	  | 5 -> green5
	  | 6 -> green6
	  | 7 -> green7
	  | 8 -> green8
	  | _ -> green9
      in
      set_color c;
      fill_circle (xx+scale_x) (yy+scale_y) (scale_x/2);
    end
;;

let show_world () =
  open_graph " 995x890";
  for y=0 to !world_height - 1 do
    for x=0 to !world_width - 1 do
      show_cell x y
    done
  done

let show_cell_at (x,y) = show_cell x y


let show_ant a =
  let (x,y) = a.position in
  let (xx,yy) = gra_pos x y in
  let c = match a.color with
    | Black -> black
    | Red -> red
  in
  set_color c;
  fill_circle (xx+scale_x) (yy+scale_y) (scale_x);
  if a.has_food then
    begin
      set_color green;
      fill_circle (xx+scale_x) (yy+scale_y) (scale_x/2);
    end;
  let w = wait_next_event [Poll ; Key_pressed ] in
  if w.keypressed then
    match w.key with
      |	' ' ->
	  begin
	    try while true do
	      let w = wait_next_event [ Key_pressed ] in 
	      match w.key with 
		| 'c' -> raise Exit
		| 'q' -> exit 0
		| _ -> ()
	    done 
	    with Exit -> ()
	  end
      | 'q' -> exit 0
      | _ -> ()
;;

(* subsection Neurology *)

let red_brain = ref [||]
let black_brain = ref [||]

let get_instruction c (s:state) : instruction =
  (* get the instruction for state s of ant color c *)
  match c with
    | Red -> !red_brain.(s)
    | Black -> !black_brain.(s)

(* subsection Neuro-Cartography *)

let verbose = ref false

let read_brain filename =
  let ch = open_in filename in
  let buf = Lexing.from_channel ch in
  let instrs = Ant_parser.instrs_eof Ant_lexer.token buf in
  close_in ch;
  if !verbose then eprintf "%d instructions read@." (List.length instrs);
  Array.of_list instrs
  
let dump_brain n b =
  for i=0 to n do
    let ins = b.(i) in
    match ins with
      | Sense(sense_dir, st1, st2, condition) ->
	  eprintf "Sense(...)@."
      | Mark(marker, st) ->
	  eprintf "Mark(...)@."
      | Unmark(marker, st) ->
	  eprintf "Ummark(...)@."
      | PickUp(st1, st2) ->
	  eprintf "PickUp(...)@."
      | Drop(st) ->
	  eprintf "Drop(...)@."
      | Turn(left_or_right, st) ->
	  eprintf "Turn(...)@."
      | Move(st1, st2) ->
	  eprintf "Move(...)@."
      | Flip(n,st1,st2) ->
	  eprintf "Flip %d %d %d@." n st1 st2
  done

(* subsection Martial Arts *)

let adjacent_ants (p:pos) c : int =
  let n = ref 0 in
  for d = 0 to 5 do
    let cel = adjacent_cell p  d in
    if some_ant_is_at cel && (ant_at cel).color = c 
    then incr n
  done;
  !n

let check_for_surrounded_ant_at (p:pos) =
  if some_ant_is_at p then 
    let a = ant_at p in
    if adjacent_ants p (other_color a.color) >= 5 then 
      begin
	kill_ant_at p;
	show_cell_at p;
	a.is_alive <- false;
	set_food_at p (food_at p + 3 + (if a.has_food then 1 else 0))
      end

let check_for_surrounded_ants (p:pos) =
  check_for_surrounded_ant_at p;
  for d = 0 to 5 do
    check_for_surrounded_ant_at (adjacent_cell p d)
  done

(* subsection Number Theory *)

let seed = ref 0

let next_seed () =
  seed := !seed * 22695477 + 1

let randomint (n:int) : int = 
  next_seed ();
  let x = (!seed lsr 16) land 16383 in
  x mod n

let randomseed n =
  seed := n;
  next_seed ();
  next_seed ();
  next_seed ()

let debug_random = false  

let _ =
  if debug_random then
    begin
      randomseed 12345;
      for i=0 to 99 do
	printf "%d, " (randomint 16384);
      done;
      printf "@."
    end


(* subsection Kinetics *)

let step (id:int) =
  if ant_is_alive id then
    let p = find_ant id in
    let a = ant_at p in
    if a.resting > 0 then
      a.resting <- a.resting - 1
    else
      match get_instruction a.color a.state with
        | Sense(sensedir, st1, st2, cond) ->
          let p' = sensed_cell p a.direction sensedir in
          let st = if cell_matches p' cond a.color then st1 else st2 in
          a.state <- st
        | Mark(i, st) ->
          set_marker_at p a.color i;
          a.state <- st
        | Unmark(i, st) ->
          clear_marker_at p a.color i;
          a.state <- st
        | PickUp(st1, st2) ->
          if a.has_food || food_at p = 0 then 
            a.state <- st2 
          else begin
            set_food_at p (food_at p - 1);
            a.has_food <- true;
            a.state <- st1
          end
        | Drop(st) ->
          if a.has_food then begin
            set_food_at p (food_at p + 1);
            a.has_food <- false
          end;
          a.state <- st
        | Turn(lr, st) -> 
          a.direction <- turn lr a.direction;
          a.state <- st
        | Move(st1, st2) ->
          let newp = adjacent_cell p a.direction in
          if rocky(newp) || some_ant_is_at(newp) then
            a.state <- st2
          else begin
            clear_ant_at p;
	    show_cell_at p;
            set_ant_at newp a;
	    show_ant a;
            a.state <- st1;
            a.resting <- 14;
            check_for_surrounded_ants(newp)
          end 
        | Flip(p, st1, st2) ->
          let st = if randomint p = 0 then st1 else st2 in
          a.state <- st

let dump = ref false

let round r =
  if !dump && r <= 10000 then dump_world r;
  if !verbose && r mod 5000 = 0 then eprintf "round %d@." r;
  for id = 0 to !max_ant_id do step id done

(* subsection Game Play and Scoring *)


let game world red black =
  let s = 12345 in
  randomseed s;
  if !dump then printf "random seed: %d@.@." s;
  let ant_hills = read_world world in
  show_world ();
  let ant_hills =
    List.sort
      (fun (x1,y1,_) (x2,y2,_) ->
	 match compare y1 y2 with
	   | 0 -> compare x1 x2
	   | n -> n)
      ant_hills
  in
  let l = List.length ant_hills in
  max_ant_id := l - 1;
  ants := Array.init l
    (fun i ->
       { id = i ; color = Red ; state = 0 ; resting = 0 ;
	 position = (0,0);
	 direction = 0 ; has_food = false ; is_alive = true });
  let _ = 
    List.fold_left
      (fun i (x,y,c) ->
	 !map.(x).(y).cell_ant <- Some i;
	 let a = !ants.(i) in
	 a.color <- c;
	 a.position <- (x,y);
	 show_ant a;
	 i+1)
      0 ant_hills
  in
  red_brain := read_brain red;
  black_brain := read_brain black;
  for i=0 to 99999 do round i done;
  (* score *)
  List.fold_left
    (fun (r,b) (x,y,c) ->
       let cell = !map.(x).(y) in
       match c with
	 | Red -> (r + cell.food,b)
	 | Black -> (r, b + cell.food))
    (0,0) ant_hills
;;
	 
(*
    
<LI>At the end, we count the number of food particles currently
 in the anthill cells of each color. The color with the most
 food wins. Remember that food being carried by an ant does not count, even if it 
 is standing on its own anthill.
</UL>
*)


let world = ref None
let red = ref None
let black = ref None

let spec =
  [ "-v", Arg.Set verbose, "verbose";
    "-dump", Arg.Set dump, "dumps 10000 first rounds of game";
    "-world", Arg.String (fun s -> world := Some s), "<world file>";
    "-red", Arg.String (fun s -> red := Some s), "<red code>";
    "-black", Arg.String (fun s -> black := Some s), "<black code>"]

let () =
  begin
    Arg.parse spec (fun _ -> ()) "gsim [options]";
    match (!world,!red,!black) with
      | Some w, Some r, Some b ->
	  let (r,b) = game w r b in
	  if !verbose then eprintf "score : %d - %d.@." r b;
	  if r > b then printf "RED@." else
	  if r < b then printf "BLACK@." else 
	    printf "DRAW@."
      | _ -> Arg.usage spec "gsim [options]"; exit 1
  end
