
(*

  a nofood gene is a function from 3 labels (lstart,lnofood,lfood)
  to instruction list

  example 1: pickup food ahead if possible (but not in home)

  lstart: Sense Ahead (lstart+1) lnofood Food 
  +1      Sense Ahead lnofood (lstart+2) Home 
  +2      Move (lstart+3) lnofood
  +3      PickUp lfood lnofood

  example 2: move randomly (never pickup food)

  lstart: Flip 3 (lstart+1) (lstart+2)
  +1      Move lnofood (lstart+2)
  +2      Flip 2 (lstart+3) (lstart+4)
  +3      Turn Left lnofood
  +4      Turn Right lnofood

  a food gene is similar but with labels (lstart,ldrop, lnodrop)

  example 3: drop food home if possible
  
  lstart: Sense Ahead (lstart+3) (lstart+1) Home 
  +1      Move (lstart+2) lnodrop
  +2      Drop ldrop
  +3      Sense LeftAhead (lstart+5) (lstart+4) Home 
  +4      Turn Left (lstart+1)
  +5      Sense RightAhead (lstart+6) lnodrop Home 
  +6      Turn Right (lstart+1)

  example 4: move randomly can be used too

*)

(*

a genome is made of two list of genes with weight, one for food gene
and one for nofood genes. 

  nofoodgenes =  n1,g1 ; ... ; nk, gk
  foodgenes =  m1,h1 ; ... ; ml, hl

The corresponding program is

  start : 
      decision tree for nofoodgenes
      gi with lnofood = 0 and lfood = gotfood

  gotfood:

      decision tree for foodgenes
      hj with ldropped = start and lnodrop = gotfood

*)


let start = ref 0;;
let nofoodfound = ref 0;;
let gotfood = ref 0;;
let gotfoodbutnothome = ref 0;;

type abs_state =
  | Fix of int
  | Ref of int ref

open Instr;;

type abs_instr =
  | ASense of sense_dir * abs_state * abs_state * condition  
  | AMark of marker * abs_state
  | AUnmark of marker * abs_state
  | APickUp of abs_state * abs_state
  | ADrop of abs_state
  | ATurn of left_or_right * abs_state
  | AMove of abs_state * abs_state
  | AFlip of int * abs_state * abs_state
;;



(*

  example 1: pickup food ahead if possible (but not in home)

  lstart: Sense Ahead (lstart+1) lnofood Food 
  +1      Sense Ahead lnofood (lstart+2) Home 
  +2      Move (lstart+3) lnofood
  +3      PickUp lfood lnofood

*)

let get_food_ahead lstart =
  [ ASense(Ahead,Fix(lstart+1),Ref(start),Food) ;
    ASense(Ahead,Ref(start),Fix(lstart+2),Home) ;
    AMove(Fix(lstart+3),Ref(start)) ;
    APickUp(Ref(gotfood),Ref(start)) ]

(*
  example 2: move randomly (never pickup food)

  lstart: Flip 3 (lstart+1) (lstart+2)
  +1      Move lnofood (lstart+2)
  +2      Flip 2 (lstart+3) (lstart+4)
  +3      Turn Left lnofood
  +4      Turn Right lnofood

*)

let move_randomly return lstart =
  [ AFlip(3,Fix(lstart+1),Fix(lstart+2)) ;
    AMove(Ref(return),Fix(lstart+2)) ;
    AFlip(2,Fix(lstart+3),Fix(lstart+4)) ;
    ATurn(Left,Ref(return)) ;
    ATurn(Right,Ref(return)) ]



(*

  example 3: drop food home if possible
  
  lstart: Sense Ahead (lstart+1) (lstart+3) Home 
  +1      Move (lstart+2) lnodrop
  +2      Drop ldrop
  +3      Sense LeftAhead (lstart+4) (lstart+5) Home 
  +4      Turn Left (lstart+1)
  +5      Sense RightAhead (lstart+6) lnodrop Home 
  +6      Turn Right (lstart+1)

*)

let drop_food_home lstart =
  [ ASense(Here,Fix(lstart+1),Fix(lstart+2), Home) ;
    ADrop(Ref(start)) ;
    ASense(Ahead,Fix(lstart+3),Fix(lstart+4), Home) ;
    AMove(Fix(lstart+1),Fix(lstart+8)) ;
    ASense(LeftAhead, Fix(lstart+5), Fix(lstart+6), Home) ;
    ATurn(Left,Fix(lstart+3)) ;
    ASense(RightAhead,Fix(lstart+7),Ref(gotfoodbutnothome), Home) ;
    ATurn(Right,Fix(lstart+3));
    AFlip(2,Fix(lstart+9),Fix(lstart+11)) ;
    ATurn(Left,Fix(lstart+10)) ;
    ATurn(Left,Fix(lstart)) ;
    ATurn(Right,Fix(lstart+12)) ;
    ATurn(Right,Fix(lstart)) ;
  ]
;;


let check_food lstart =
  [ ASense(Ahead,Fix(lstart+1),Fix(lstart+4), Food) ;
    ASense(Ahead,Fix(lstart+4),Fix(lstart+2),Home) ;
    AMove(Fix(lstart+3),Fix(lstart+4)) ;
    APickUp(Ref(gotfood),Fix(lstart+4)) ;
    ASense(LeftAhead, Fix(lstart+5), Fix(lstart+7), Food) ;
    ASense(Ahead,Fix(lstart+7),Fix(lstart+6),Home) ;
    ATurn(Left,Fix(lstart+2)) ;
    ASense(RightAhead,Fix(lstart+8),Ref(nofoodfound), Food) ;
    ASense(Ahead,Ref(nofoodfound),Fix(lstart+9),Home) ;
    ATurn(Right,Fix(lstart+2)) ]
;;

let rec compile_tree next tree =
  match tree with
    | Decision_tree.Leaf x -> 
	let l = x next in
	(l,next+List.length l)
    | Decision_tree.Flip(n,t1,t2) ->
	let ct2,n2 = compile_tree (next+1) t2 in
	let ct1,n1 = compile_tree n2 t1 in
	( AFlip(n,Fix(next+1),Fix(n2)) :: ct2 @ ct1 , n1)
;;

let rel = function
  | Fix n -> n
  | Ref n -> !n
;;

let relocate i = 
  match i with
  | ASense(sd,s1,s2,c) -> Sense(sd,rel s1,rel s2,c)
  | AMark(m,s) -> Mark(m,rel s)
  | AUnmark(m,s) -> Unmark(m,rel s)
  | APickUp(s1,s2) -> PickUp(rel s1, rel s2)
  | ADrop(s) -> Drop(rel s)
  | ATurn(d,s) -> Turn(d,rel s)
  | AMove(s1,s2) -> Move(rel s1, rel s2)
  | AFlip(n,s1,s2) -> Flip(n,rel s1, rel s2)
;;
	
let compile nofoodgenes foodgenes =
  let nofoodt = Decision_tree.build nofoodgenes
  and foodt = Decision_tree.build foodgenes
  in
  start := 0;
  let check_for_food = check_food 0 in
  nofoodfound := List.length check_for_food;
  let instr_nofood,next = compile_tree !nofoodfound nofoodt in
  gotfood := next;
  let check_food_drop = drop_food_home next in
  gotfoodbutnothome := next + List.length check_food_drop;
  let instr_food,next = compile_tree !gotfoodbutnothome foodt in
  List.map relocate (check_for_food@instr_nofood@check_food_drop@instr_food)

  
  
let write_ant file l =
  let ch = open_out ("ants/"^file^".ant") in
  Instr.output ch l;
  close_out ch
;;
    
let test1 =
  compile 
    [ (2,move_randomly start) ; (1,get_food_ahead) ]
    [ (2,move_randomly gotfood) ; (1,drop_food_home) ]
;;

write_ant "test1" test1;;


(*

  follow_mark n : follow mark n if possible
  
*)

let follow_mark return n lstart =
  [ ASense(Ahead,Fix(lstart+1),Fix(lstart+2),Marker(n)) ;
    AMove(Ref(return),Ref(return)) ;
    ASense(LeftAhead,Fix(lstart+3),Fix(lstart+4),Marker(n)) ;
    ATurn(Left,Fix(lstart+1)) ;
    ASense(RightAhead,Fix(lstart+5),Ref(return),Marker(n)) ;
    ATurn(Right,Fix(lstart+1)) ]
;;

let put_mark return n lstart =
  [ AMark(n,Ref(return)) ]
;;



let test2 =
  compile 
    [ (2,move_randomly start) ; (1,get_food_ahead) ;
      (1,follow_mark start 0) ; (1,put_mark start 3)
    ]
    [ (2,move_randomly gotfood) ; (1,drop_food_home);
      (1,follow_mark gotfood 3) ; (1,put_mark gotfood 0) ]
;;

write_ant "test2" test2;;

(*

  move n steps

*)

let rec make_list lstart return i l =
  if i=0 then l
  else make_list lstart return (i-1) (AMove(Fix(lstart+i),Ref(return)) :: l)

let move_steps return n lstart =
  make_list lstart return (n-1) [ AMove(Ref(return),Ref(return)) ]
;;

let test3 =
  compile 
    [ (2,move_randomly start) ; (1,get_food_ahead) ;
      (1,follow_mark start 0) ; (1,put_mark start 3) ;
      (1,move_steps start 10) 
    ]
    [ (2,move_randomly gotfood) ; (1,drop_food_home);
      (1,follow_mark gotfood 3) ; (1,put_mark gotfood 0) ;
      (1,move_steps gotfood 10) 
    ]
;;

write_ant "test3" test3;;

let test4 =
  compile 
    [ (2,move_randomly start) ; (1,get_food_ahead) ;
      (1,follow_mark start 0) ; (1,put_mark start 3) ;
      (1,move_steps start 10) 
    ]
    [ (2,move_randomly gotfood) ; 
      (1,follow_mark gotfood 3) ; (1,put_mark gotfood 0) ;
      (1,move_steps gotfood 10) 
    ]
;;

write_ant "test4" test4;;

let test6 =
  compile 
    [ (2,move_randomly start) ; 
      (1,follow_mark start 0) ; (1,put_mark start 3) ;
      (1,move_steps start 10) 
    ]
    [ (1,move_randomly gotfood) ; 
      (1,follow_mark gotfood 3) ; (1,put_mark gotfood 0) ;
      (1,move_steps gotfood 5) 
    ]
;;

write_ant "test6" test6;;

let attack lstart =
  [ ASense(Ahead,Ref(start),Fix(lstart+1), Foe) ;
    ASense(LeftAhead, Fix(lstart+2), Fix(lstart+3), Foe) ;
    ATurn(Left,Ref(start)) ;
    ASense(RightAhead,Fix(lstart+4),Ref(start), Foe) ;
    ATurn(Right,Ref(start)) ]
      

let test7 =
  compile 
    [ (2,move_randomly start) ; 
      (1,follow_mark start 0) ; (1,put_mark start 3) ;
      (1,move_steps start 10) ; (5,attack) 
    ]
    [ (1,move_randomly gotfood) ; 
      (1,follow_mark gotfood 3) ; (1,put_mark gotfood 0) ;
      (1,move_steps gotfood 5) 
    ]
;;

write_ant "test7" test7;;

